
#include <Timer.h>

#define BIT0 0x01
#define BIT1 0x02
#define BIT2 0x04
#define BIT3 0x08
#define BIT4 0x10
#define BIT5 0x20
#define BIT6 0x40
#define BIT7 0x80

#define READ_BIT    		BIT7
#define PKT_MAX_LENGTH	28

#define WITH_ln			0
#define NONE_ln			1

#define I2C_W_NONE_ADDR		0
#define I2C_W_WITH_ADDR		1

//=========== Pin Define ==========================
//UART_1 For Communication with RF_Link_UART_Root
#define UART_1_RX        19
#define UART_1_TX        18

#define O_N_CEB          9

#define O_Root_Mode      6  //Low=CMD, High=Normal
#define I_ID_1           5  //INPUT_PULLUP
#define I_ID_0           4  //INPUT_PULLUP

//UART_0 For Display On PC
#define UART_0_TX        1
#define UART_0_RX        0

//=================================================


Timer Mode_1_Timer;
int tickEvent;

uint8_t DebounceCnt = 0;

uint8_t Teset_Step = 0;
uint8_t Last_Teset_Step = 0;

uint8_t TX_Buffer[256];	//Data to RF_Link_UART_Root
uint8_t RX_Buffer[256]; //Data from RF_Link_UART_Root

uint8_t ID_SEL = 0;
uint8_t NewDevNum;

bool ID_Change_Flag;

uint8_t LCD_Cursor = 0;
uint8_t EE_ADDR = 0;
uint8_t DataLength = 0;

#define EE_24C04_I2C_ID			0x50
#define LCD1602_I2C_ID			0x27

//========================================================= LCD1602 Parameter define =====
#define LCD_CLEARDISPLAY    0x01
#define LCD_RETURNHOME      0x02
#define LCD_ENTRYMODESET    0x04  //Value in Display_MODE
//Display_MODE
#define LCD_ENTRYRIGHT          0x00  //text that flows Right to Left
#define LCD_ENTRYLEFT           0x02  //text that flows Left to Right
#define LCD_ENTRYSHIFTINCREMENT 0x01  //right justify text from the cursor
#define LCD_ENTRYSHIFTDECREMENT 0x00  //left justify text from the cursor

#define LCD_DISPLAYCONTROL  0x08  //Value in Display_CTRL
//Display_CTRL
#define LCD_DISPLAYON           0x04
#define LCD_DISPLAYOFF          0x00
#define LCD_CURSORON            0x02  //underline cursor on
#define LCD_CURSOROFF           0x00
#define LCD_BLINKON             0x01  //cursor blinking on
#define LCD_BLINKOFF            0x00

#define LCD_CURSORSHIFT     0x10
// flags for display/cursor shift
#define LCD_DISPLAYMOVE         0x08
#define LCD_CURSORMOVE          0x00
#define LCD_MOVERIGHT           0x04
#define LCD_MOVELEFT            0x00

#define LCD_FUNCTIONSET     0x20  //Value in Display_FUNCTION
//Display_FUNCTION
#define LCD_8BITMODE            0x10
#define LCD_4BITMODE            0x00
#define LCD_2LINE               0x08
#define LCD_1LINE               0x00
#define LCD_5x10DOTS            0x04
#define LCD_5x8DOTS             0x00

#define LCD_SETCGRAMADDR    0x40
#define LCD_SETDDRAMADDR    0x80

#define LCD_BACKLIGHT       0x08

#define En    0x04  // Enable bit
#define Rw    0x02  // Read/Write bit
#define Rs    0x01  // Register select bit
//=========================================================
#define LCD1602_Clear()   LCD1602_Write_Command(LCD_CLEARDISPLAY, 0)
#define LCD1602_Home()    LCD1602_Write_Command(LCD_RETURNHOME, 0)

//========================================================================================

char FW_VER[] = "190614_01";

uint8_t DevNum_Table[4] = //(DevNum Range=0~63)
{
  0,
  1,
  2,
  3
};

uint8_t I2C_Slave_ID_Table[4] =
{
  EE_24C04_I2C_ID,	//0x50
  LCD1602_I2C_ID,		//0x27
  LCD1602_I2C_ID,		//0x27
  LCD1602_I2C_ID		//0x27
};

/*---------------------------------------------------------------------------------------------------------*/
/* Function   : LCD_expanderWrite                                                                          */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void LCD_expanderWrite(uint8_t Data)
{
  TX_Buffer[0] = (Data | LCD_BACKLIGHT); //Back Light always on
  
  //Write without I2C_ADDR
  Send_Data_To_I2C_Device(I2C_Slave_ID_Table[ID_SEL], DevNum_Table[ID_SEL], 1, &TX_Buffer[0], I2C_W_NONE_ADDR);
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : LCD_write4bits                                                                             */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void LCD_write4bits(uint8_t Data)
{
  LCD_expanderWrite(Data);

  LCD_expanderWrite(Data | En); // En high

  LCD_expanderWrite((Data & ~(En))); // En low
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : LCD1602_Write_Command                                                                      */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void LCD1602_Write_Command(uint8_t CMD, uint8_t Mode)
{
  //write either command or data
  LCD_write4bits((CMD & 0xF0) | Mode);    //High Nibble
  LCD_write4bits(((CMD << 4) & 0xF0) | Mode); //Low Nibble
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : LCD1602_SetCursor                                                                          */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void LCD1602_SetCursor(uint8_t Col, uint8_t Row)
{
  (Row) ? (Row = 0x40) : (Row = 0);

  LCD1602_Write_Command(LCD_SETDDRAMADDR | (Col + Row), 0);

  (Row) ? (LCD_Cursor = Col + 16) : (LCD_Cursor = Col);
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : Init_LCD1602                                                                               */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void Init_LCD1602(void)
{
  uint8_t i;

  //Pull both RS and R/W low to begin commands
  LCD_expanderWrite(0);

  //Put the LCD into 4 bit mode, start in 8bit mode, try to set 4 bit mode
  for (i = 0; i < 3; i++)
    LCD_write4bits(0x03 << 4);

  // finally, set to 4-bit interface
  LCD_write4bits(0x02 << 4);

  //Set lines, font size, etc.
  LCD1602_Write_Command(LCD_FUNCTIONSET | (LCD_4BITMODE | LCD_2LINE | LCD_5x8DOTS), 0);

  //Turn the display on with no cursor or blinking default
  LCD1602_Write_Command(LCD_DISPLAYCONTROL | (LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF), 0);

  //Initialize to default text direction (for roman languages)
  LCD1602_Write_Command(LCD_ENTRYMODESET | (LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT), 0);

  LCD1602_Clear();  //Clear display

  LCD_Cursor = 0;
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : CharToHex                                                                                  */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
uint8_t CharToHex(char ch)
{
  uint8_t RetData = 255;

  switch (ch)
  {
    case '0':
      RetData = 0;
      break;

    case '1':
      RetData = 1;
      break;

    case '2':
      RetData = 2;
      break;

    case '3':
      RetData = 3;
      break;

    case '4':
      RetData = 4;
      break;

    case '5':
      RetData = 5;
      break;

    case '6':
      RetData = 6;
      break;

    case '7':
      RetData = 7;
      break;

    case '8':
      RetData = 8;
      break;

    case '9':
      RetData = 9;
      break;

    case 'a':
    case 'A':
      RetData = 10;
      break;

    case 'b':
    case 'B':
      RetData = 11;
      break;

    case 'c':
    case 'C':
      RetData = 12;
      break;

    case 'd':
    case 'D':
      RetData = 13;
      break;

    case 'e':
    case 'E':
      RetData = 14;
      break;

    case 'f':
    case 'F':
      RetData = 15;
      break;
  }
  return RetData;
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : StringToNum                                                                                 */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
bool StringToNum(uint8_t* pBuf, uint8_t BufLength)
{
  uint8_t i, HexTemp;

  //2String--->1Byte Num
  //EX: String-->2F to Num=0x2F

  for (i = 0; i < BufLength; i++)
  {
    HexTemp = CharToHex(pBuf[i]);

    if (HexTemp >= 16)
    {
      Serial.println("InValid HEX !!");
      return false;
    }
    else
    {
      if (!(i % 2)) //High_Nibble
        pBuf[i / 2] = (HexTemp << 4); //HexTemp*16
      else       //Low_Nibble
        pBuf[i / 2] += HexTemp;
    }
  }

  if (BufLength % 2)    //Last String is High_Nibble
    pBuf[(i - 1) / 2] >>= 4;

  return true;
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : UART0_Clr_Data                                                                             */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void UART0_Clr_Data(void)
{
  //Clr all Data from PC
  do
  {
    Serial.read();
  }
  while (Serial.available());
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : UART0_Print_Hex                                                                            */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void UART0_Print_Hex(uint8_t* pBuf, uint8_t Length, bool none_ln_flag)
{
  uint8_t i;

  for (i = 0; i < Length; i++)
  {
    Serial.print("0x");
    if (pBuf[i] <= 0x0F)
      Serial.print("0");

    Serial.print(pBuf[i], HEX);

    if (none_ln_flag)
      return;
    else if (i < (Length - 1))
    {
      if (!((i + 1) % 16))
        Serial.println("");
      else
        Serial.print(",");
    }
    else
      Serial.println("");
  }
  Serial.println("");
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : UART1_Clr_Data                                                                             */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void UART1_Clr_Data(void)
{
  //Clr all Data from RF_Link_UART_Root
  do
  {
    Serial1.read();
  }
  while (Serial1.available());
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : Send_RF_Link_CMD                                                                           */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void Send_RF_Link_CMD(uint8_t CMD, uint8_t Length)
{
  //use UART_1 Send to RF_Link_UART_Root
  digitalWrite(O_Root_Mode, LOW); //Low=CMD_Mode

  //;---------------------------------------------
  //Byte0=CMD
  //  Bit7= R/W (0=Write, 1=Read)
  //  Bit6=Reserved
  //  B[5:0]=Device_ID
  Serial1.write(CMD);

  //Byte1=Normal_Mode_DataLength
  Serial1.write(Length);

  Serial1.flush();  //Wait for the transmission of outgoing serial data to complete
  delayMicroseconds(50);
  //;---------------------------------------------

  digitalWrite(O_Root_Mode, HIGH); //High=Normal_Mode

  delayMicroseconds(150);
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : Send_Data_To_I2C_Device                                                                    */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void Send_Data_To_I2C_Device(uint8_t I2C_Slave_ID, uint8_t DevNum, uint8_t DataLength, uint8_t* pData, bool EE_ADDR_Flag)
{
  //DataLength not include I2C_Slave_ID & EE_ADDR !!!

  if (EE_ADDR_Flag)
    Send_RF_Link_CMD(DevNum, (DataLength + 1));	 //Length+1 for EE_ADDR
  else
    Send_RF_Link_CMD(DevNum, DataLength);

  Serial1.write((I2C_Slave_ID << 1)); 	 		 //I2C Slave ID+W

  if (EE_ADDR_Flag)
    Serial1.write(EE_ADDR);

  if (DataLength)
    Serial1.write(&pData[0], DataLength);

  Serial1.flush();  //Wait for the transmission of outgoing serial data to complete
  delayMicroseconds(50);
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : Read_I2C_Device                                                                            */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void Read_I2C_Device(uint8_t I2C_Slave_ID, uint8_t DevNum, uint8_t ReadLength, bool EE_ADDR_Flag)
{
  if (EE_ADDR_Flag)
    Send_Data_To_I2C_Device(I2C_Slave_ID, DevNum, 0, &TX_Buffer[0], I2C_W_WITH_ADDR); 	//Do I2C Write for Shift to Last EE_ADDR

  //Send I2C Read Request
  Send_RF_Link_CMD((DevNum | READ_BIT), ReadLength);

  Serial1.write(((I2C_Slave_ID << 1) | BIT0)); //I2C Slave ID+R
  Serial1.flush();  //Wait for the transmission of outgoing serial data to complete
  delayMicroseconds(50);
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : DeviceChange                                                                               */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void DeviceChange(void)
{
  ID_SEL = NewDevNum;
  ID_Change_Flag = 0;
  DebounceCnt = 0;

  Teset_Step = 0;
  Last_Teset_Step = 0xFF;

  LCD_Cursor = 0;
  EE_ADDR = 0;
  DataLength = 0;

  Serial.println("");
  Serial.print("===== Device");
  Serial.print(DevNum_Table[ID_SEL]);

  if (!(ID_SEL))	//Device 0
    Serial.println(" 24C04EE =====");
  else					//Device 1~3
  {
    Serial.println(" LCD1602 =====");
    Serial.println("Key in Data will show in LCD");
  }

  Send_RF_Link_CMD(DevNum_Table[ID_SEL], 0);
  delay(50);  //delay 50ms for RFLink Root Sync to New Device

  UART0_Clr_Data(); //Clr all Data from PC
  UART1_Clr_Data(); //Clr all Data from RF_Link_UART_Root

  if (ID_SEL)
    Init_LCD1602();
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : Timer_1ms                                                                                  */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void Timer_1ms(void* context)
{
  //  PIN5   PIN4
  //I_ID_1, I_ID_0 default INPUT_PULLUP
  //   1      1     Devcie0
  //   1      0     Devcie1
  //   0      1     Devcie2
  //   0      0     Devcie3
  if (digitalRead(I_ID_0))
    NewDevNum = 0;
  else
    NewDevNum = BIT0;

  if (digitalRead(I_ID_1))
    NewDevNum &= ~(BIT1);
  else
    NewDevNum |= BIT1;

  if (NewDevNum != ID_SEL)
  {
    if (DebounceCnt < 150)
      DebounceCnt++;
    else
      ID_Change_Flag = 1;
  }
  else
    DebounceCnt = 0;
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : setup                                                                                      */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void setup()
{
  // UART_0 For Display On PC
  Serial.begin(115200);

  tickEvent = Mode_1_Timer.every(1, Timer_1ms, (void*)1);

  pinMode(I_ID_1, INPUT_PULLUP);      //5
  pinMode(I_ID_0, INPUT_PULLUP);      //4

  //--------------------------------------
  pinMode(O_N_CEB, OUTPUT);           //9
  digitalWrite(O_N_CEB, HIGH); 				//High for turn off RF_Link_UART_Root PWR

  //all output low of connect with RF_Link_UART_Root
  pinMode(UART_1_RX, OUTPUT);         //19
  digitalWrite(UART_1_RX, LOW);

  pinMode(UART_1_TX, OUTPUT);         //18
  digitalWrite(UART_1_TX, LOW);

  pinMode(O_Root_Mode, OUTPUT);       //6
  digitalWrite(O_Root_Mode, LOW);

  delay(200);
  //--------------------------------------

  digitalWrite(O_N_CEB, LOW);         //9  Low for turn on RF_Link_UART_Root PWR

  // UART_1 For Communication with RF_Link_UART_Root
  Serial1.begin(19200);

  pinMode(O_Root_Mode, OUTPUT);       //6
  digitalWrite(O_Root_Mode, HIGH);

  delay(1000);  //delay for RF_Link_UART_Root POR_Initial
  //--------------------------------------


  Serial.println("=========================================");
  Serial.print("Ver=");
  Serial.println(FW_VER);

  ID_Change_Flag = 1;
}
/*---------------------------------------------------------------------------------------------------------*/
/* Function   : loop                                                                                       */
/* Parameters : None                                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void loop()
{
  uint8_t Length;

  Mode_1_Timer.update();

  if (ID_Change_Flag)
    DeviceChange();

  if (ID_SEL)	//LCD_1602
  {
    UART1_Clr_Data(); //Clr all Data from RF_Link_UART_Root

    if (Serial.available()) //Data From UART_0 Now !!!
    {
      LCD1602_Write_Command(Serial.read(), Rs);
      LCD_Cursor++;

      if (LCD_Cursor == 16)
        LCD1602_SetCursor(0, 1);
      else if (LCD_Cursor == 32)
        LCD1602_SetCursor(0, 0);
    }
  }
  else //EE_24C04
  {
    switch (Teset_Step)
    {
      case 0: //================================================================================== Wait for R/W CMD input
        if (Last_Teset_Step != Teset_Step)
        {
          Serial.println("Please Press R or W for EEPROM Proc");
          Last_Teset_Step = Teset_Step;
        }

        UART1_Clr_Data(); //Clr all Data from RF_Link_UART_Root

        if (Serial.available()) //Data From UART_0 Now !!!
        {
          delay(10); //delay for UART Data in
          Length = Serial.available();

          if (Length >= 2)
          {
            Serial.println("InValid CMD !!!");
            UART0_Clr_Data();  //Clr all Data from PC
          }
          else
          {
            Serial.readBytes(&TX_Buffer[0], 1);

            if ((TX_Buffer[0] == 'W') || (TX_Buffer[0] == 'w'))
              Teset_Step = 1;
            else if ((TX_Buffer[0] == 'R') || (TX_Buffer[0] == 'r'))
              Teset_Step = 3;
            else
              Serial.println("InValid CMD !!!");
          }
        }
        break;

      case 1: //================================================================================== Wait for Write EE_ADDR input
      case 3: //================================================================================== Wait for Read EE_ADDR input
        if (Last_Teset_Step != Teset_Step)
        {
          Serial.print("Current EERROM_ADDR(HEX)=");
          UART0_Print_Hex(&EE_ADDR, 1, NONE_ln);

          if (Teset_Step == 1)
            Serial.println(", Which EERROM_ADDR for Write (HEX-->00~FF)");
          else	//Teset_Step==3
            Serial.println(", Which EERROM_ADDR for Read (HEX-->00~FF)");

          Last_Teset_Step = Teset_Step;
        }

        UART1_Clr_Data(); //Clr all Data from RF_Link_UART_Root

        if (Serial.available()) //Data From UART_0 Now !!!
        {
          delay(10); //delay for UART Data in
          Length = Serial.available();

          if (Length >= 3)
          {
            Serial.println("EERROM_ADDR over 0xFF !");
            UART0_Clr_Data();  //Clr all Data from PC
          }
          else
          {
            Serial.readBytes(&TX_Buffer[0], Length);

            if (StringToNum(&TX_Buffer[0], Length))
            {
              EE_ADDR = TX_Buffer[0];
              Teset_Step++;
            }
          }
        }
        break;

      case 2: //================================================================================== Wait for EE_Data input
        if (Last_Teset_Step != Teset_Step)
        {
          Serial.println("Please Input Data For Write (Max 16Bytes), * for write EE_ADDR only");
          Last_Teset_Step = Teset_Step;
        }

        UART1_Clr_Data(); //Clr all Data from RF_Link_UART_Root

        if (Serial.available()) //Data From UART_0 Now !!!
        {
          delay(10); //delay for UART Data in
          Length = Serial.available();

          if (Length >= 33)
          {
            Serial.println("Max 16Bytes/Page For Write!");
            UART0_Clr_Data();  //Clr all Data from PC
          }
          else
          {
            DataLength = Length / 2; //WriteLength
            Serial.readBytes(&TX_Buffer[0], Length);

            if ((Length == 1) && (TX_Buffer[0] == '*'))
              goto WRITE_EE_ADDR_ONLY;
            else
            {
              if (!(StringToNum(&TX_Buffer[0], Length)))
                return;

              if (Length % 2)
                DataLength++;
            }
          }

WRITE_EE_ADDR_ONLY:
          Teset_Step = 0;
          Serial.print("EERROM_ADDR(HEX)=");
          UART0_Print_Hex(&EE_ADDR, 1, NONE_ln);
          Serial.println("");

          if (DataLength)
          {
            Serial.print("EERROM_DATA(HEX)=");
            UART0_Print_Hex(&TX_Buffer[0], DataLength, WITH_ln);
          }
          Serial.println("-----------------------------------");

          //With Do I2C Write for Shift to Last EE_ADDR
          Send_Data_To_I2C_Device(I2C_Slave_ID_Table[ID_SEL], DevNum_Table[ID_SEL], DataLength, &TX_Buffer[0], I2C_W_WITH_ADDR);

          if (DataLength != 16)
            EE_ADDR += DataLength;
        }
        break;

      case 4: //================================================================================== Wait for Read EE_Length input
        if (Last_Teset_Step != Teset_Step)
        {
          Serial.println("Please Input Read Length, Max 27ytes(HEX-->01~1B)");
          Last_Teset_Step = Teset_Step;
        }

        UART1_Clr_Data(); //Clr all Data from RF_Link_UART_Root

        if (Serial.available()) //Data From UART_0 Now !!!
        {
          delay(10); //delay for UART Data in
          Length = Serial.available();

          if (Length >= 3)
          {
            Serial.println("Read Length over 0xFF !");
            UART0_Clr_Data();  //Clr all Data from PC
          }
          else
          {
            Serial.readBytes(&TX_Buffer[0], Length);

            if (!(StringToNum(&TX_Buffer[0], Length)))
              return;

            DataLength = TX_Buffer[0];

            if (DataLength == 0)
              Serial.println("Read Length can't be 0 !!!");
            else if (DataLength >= 28)
              Serial.println("Read Length over 27 !!!");
            else
            {
              Teset_Step = 0;

              Serial.print("EERROM_ADDR(HEX)=");
              UART0_Print_Hex(&EE_ADDR, 1, NONE_ln);
              Serial.println("");

              Serial.print("Read_Length(HEX)=");
              UART0_Print_Hex(&DataLength, 1, WITH_ln);

              //With Do I2C Write for Shift to Last EE_ADDR
              Read_I2C_Device(I2C_Slave_ID_Table[ID_SEL], DevNum_Table[ID_SEL], DataLength, I2C_W_WITH_ADDR);

              Serial1.setTimeout(1000);  //Max Wait 1sec
              Length = Serial1.readBytes(&RX_Buffer[0], DataLength);

              if (Length == 0)
                Serial.println("1sec TimeOut !!");
              else
              {
                EE_ADDR += DataLength;
                UART0_Print_Hex(&RX_Buffer[0], DataLength, WITH_ln);
              }
              Serial.println("-----------------------------------");
            }
          }
        }
        break;
    }
  }
}
